# CHECK BIC_CHART RAM USAGE LOG FOR ABNORMAL RAM USAGE EVENTS
# AND IF AN EVENT IS FOUND, OPTIONALLY...
# 1) ADD ENTRY TO THE BLUE IRIS LOGFILE
# 2) PUSH NOTIFICATION (VIA PUSHOVER API)
# 3) RESTART THE BLUE IRIS SERVICE (OR REBOOT THE BI SERVER)
#
# NOTE: THIS SCRIPT USES A RAM USAGE LOG CREATED BY BLUE IRIS ACTION SETS
#		Example...
#		Timestamp,SoftwareUsedRamBytes,Uptime,RamPct,Tag
#		2023/01/01_02:00:01,906797056,0:13:58:29,66%,"periodic"
#		2023/01/01_02:06:00,1680392192,0:14:04:28,71%,"periodic"
#		2023/01/01_02:12:00,2056347648,0:14:10:28,73%,"periodic"
#		2023/01/01_02:18:03,2370572288,0:14:16:31,75%,"periodic"
#
# Version: 2.1
# Created by jaydeel at ipcamtalk.com
#

# Updates:
# V2.1 - 2023-05-08 - fixed bug overwriting pushover notification
#                   - additional real-life test examples
# V2   - 2023-04-20 - improved demo mode handling
#                   - additional real-life test examples
# V1   - 2023-01-09 - initial release
#
# THIS SCRIPT CAN DETECT TWO TYPES OF ABNORMAL RAM USAGE EVENTS:
# 1) CONT-LEAK -- RAM usage continuously increases, ultimately maxing out at the system's RAM
# 2) RAM-STEP  -- RAM usage jumps to some multiple of the 'normal' level, then stays around this new level.
#
# USAGE:      {this script name}.ps1 [arg1]
# ARGUMENTS:  arg1 = [OPTIONAL} demo - run script in demo mode (disables logging, notifications, actions) 
#             arg1 = [OPTIONAL} restart|reboot - force selected action
#
# BEFORE USING THIS SCRIPT:
#   1. EDIT ALL VARIABLES IN 'USER SETTINGS' SECTION BELOW
#   2. COMMENT OUT UNUSED/UNNEEDED LINES
#
# HOW TO AUTOMATE THIS SCRIPT VIA TASK SCHEDULER
#   Open Task Scheduler and create an appropriately named task with two triggers and one action:
#   Task Scheduler
#   'Triggers' tab:
#       'At startup'
#       'Daily' or 'One time' (e.g., at 12:05 AM, every 6 minutes)
#       (Note: I find 'One time' to be more reliable)
#   'Actions' tab:
#       Program/script: powershell.exe
#       Add arguments: {your_powershell_scripts_path)\{script_file_name}.ps1
#

# MUST BE THE FIRST LINE IN A SCRIPT
param($p1)

#Set-StrictMode  -Version 3

# ================================================
# USER SETTINGS...

# [REQUIRED] ENTER SOURCE CSV FILE PATH...
$path = 'C:\BlueIris\www\bic_chart\_my_data'
#$path = 'C:\BlueIris\www\bic_chart\_my_data\tests'

# [REQUIRED] ENTER SOURCE CSV FILE FILENAME...
#$fname = 'ram_log.csv'
$fname = 'ram_log[1].csv'
#$fname = 'ram_leak_test_00.csv'  # no leak
#$fname = 'ram_leak_test_01.csv'  # continuous leak (7+ GB)
#$fname = 'ram_leak_test_02.csv'  # step leak (1+ GB)
#$fname = 'ram_leak_test_03.csv'  # continuous leak (4 GB)
#$fname = 'ram_leak_test_04.csv'  # continuous + step leak (12+ GB)
#$fname = 'ram_leak_test_05.csv'  # file has only 3 data rows
#$fname = 'ram_leak_test_06.csv'  # continuous leak (1+ GB)
#$fname = 'ram_leak_test_07.csv'  # continuous leak, small (200 MB)
#$fname = 'ram_leak_test_08.csv'  # continuous leak (3 GB)
#$fname = 'ram_leak_test_09.csv'  # continuous leak (1.5 GB)
#$fname = 'ram_leak_test_10.csv'  # continuous leak (3 GB)
#$fname = 'ram_leak_test_11.csv'  # continuous leak (12.5 GB)
#$fname = 'ram_leak_test_12.csv'  # continuous leak (6 GB)

# [REQUIRED] TEST FOR CONTINUOUS LEAK EVENT
# ... Note: this option is skipped if the datafiles contains less than 4 data rows
	# ENTER DELTA (IN MB) USED TO CONFIRM AN CONT-LEAK EVENT
	# ... when 3 successive RAM deltas exceed this value
	# ... recommend 50-300; default is 50
	# ... set to zero to use $ram_ratio_threshold instead
	$ram_delta_threshold_MB = 50
	# ENTER RATIO USED TO CONFIRM AN CONT-LEAK EVENT
	# ... event is confirmed when 3 successive RAM ratios exceed this value
	# ... expecting a value above 1.0
	# ... recommend 1.01 to 1.05; lower if $act_after_N_warnings > 3
	# ... ignored if $ram_delta_threshold_MB > 0
	$ram_ratio_threshold = 1.04

# [OPTIONAL] ALSO TEST FOR RAM-STEP EVENT
# ... Note: this option is skipped if the datafiles contains less than 4 data rows
	# ENTER RATIOS USED TO ID A RAM-STEP
	# ... Trigger event when the 3rd most recent ratio exceeds this value, AND...
	#     (set to zero to skip this option)
	$step_ratio_threshold = 2.0
	# ... the two most recent ratios exceed this value.
	$step_plateau_ratio_threshold = 0.95
	# [REQUIRED] ENTER MINIMUM RAM
	# ... Wait until RAM usage is above this value (in bytes) before testing for step type leaks
	# ... (needed because RAM usage is sometimes very low when rebooting Blue Iris PC)
	$step_ram_min = 700000000 

# [REQUIRED] ENTER AUTOMATED ACTION EVENTS COUNT TRIGGER
# ... wait until this many contiguous warnings before taking action
# ... applies to CONT-LEAK events only; RAM-STEP events act after single warning
# ... if you want an opportunity to inspect your server instead of permitting an automatic action
#     enter 0 (zero) to never act (e.g., 'opt out'). This applies to both types of leaks
#$act_after_N_warnings = 0  # Opt out
$act_after_N_warnings = 2

# [OPTIONAL] REBOOT BLUE IRIS SERVER INSTEAD OF DEFAULT (RESTART THE BI SERVICE)
# ... expecting $false or $true (default is $false)
$reboot_flg = $false

# [OPTIONAL] OPT-IN TO POST CUSTOM MESSAGE TO BLUE IRIS LOG
# ... expecting $false or $true (default = $false)
$post_to_BI_logfile_flg = $false

	# THE FOLLOWING ARE [REQUIRED] IF OPTING IN...

	# ENTER BLUE WEBSERVER ADDRESS & PORT
	# ... to use credentials, be sure to uncheck 'Use secure session keys and login page' 
	#     in Blue Iris settings: click Settings > Web server > Advanced
	$bi_webserver = "192.168.1.3:8600"

	# OPT-IN TO USE CREDENTIALS FOR POSTING TO BLUE IRIS LOGFILE
	# ... expecting $false or $true
	#$post_w_creds_flg = $false
	$post_w_creds_flg = $true

	# [OPTIONAL] ENTER BLUE IRIS WEBSERVER USERNAME & PASSWORD 
	# ... NOTE: USERNAME MUST BE ADMIN
	$bi_username = "YOUR_BI_USER_NAME"
	$bi_pw = "YOUR_BI_USER_PASSWORD"

# [OPTIONAL] OPT-IN TO SEND PUSHOVER NOTIFICATION
# ... expecting $false or $true (default is $false)
$send_po_notification_flg = $false

	# THE FOLLOWING ARE [REQUIRED] IF OPTING IN...

	# ENTER PUSHOVER API USER KEY
	# ... https://pushover.net/
	$po_user = "YOUR_API_USER_KEY"

	# ENTER PUSHOVER API APPLICATION TOKEN
	# ... TIP: consider adding a 'PowerShell' application to your https://pushover.net/ account
	$po_token = "YOUR_API_APPLICATION_TOKEN"

	# ENTER PUSHOVER MESSAGE PRIORITY
	# ... (integer) -2|-1|0|1|2; -2=lowest, -1=low(silent), 0=normal, 1=high-priority, 2=emergency.
	# ... see https://pushover.net/api#priority
	$po_priority = 2

	# ENTER PUSHOVER SOUND
	# ... blank ("") is default sound
	# ... see https://pushover.net/api#sounds
	#$po_sound = ""  # default sound
	#$po_sound = "siren"  # built-in sound
	$po_sound = "not-kiddin"  # custom sound

	# ENTER PUSHOVER DEVICE(S)
	# ... blank ("") sends to all devices
	# ... may be comma-delimited list.
	# ... see https://pushover.net/api
	#$po_device = ""  # all devices
	$po_device = "galaxy-s21"  # single device
	#$po_device = "galaxy-s21,iPhoneXS"  # multiple devices

	# ENTER PUSHOVER RETRY PARAMETER (SECONDS)
	# ... Required for priority=2 messages (emergency)
	# ... see https://pushover.net/api#priority
	$po_retry = 30

	# ENTER PUSHOVER EXPIRE PARAMETER (SECONDS)
	# ... Required for priority=2 messages (emergency)
	# ... see https://pushover.net/api#priority
	$po_expire = 300

# END USER SETTINGS
# ===============================================

function PostCustomBiLogEntry($msg, $level, $silent_flg) {
	# PURPOSE: Add custom message to the Blue Iris logfile
	# ARGS:    $msg        = (string) NOTE: be sure spaces are replaced with '%20'
	#          $level      = (integer) 0|1|2 (0:info, 1:warn, 2:error)
	#                         NOTE: level=2 can trigger a status alert if the 'Send only if there are new Errors' box 
	#                         is checked on the 'Status alerts' dialog (see button at bottom of the 'Blue Iris Status' (log) dialog)
	#          $silent_flg = (boolean) $true prevents default output to the terminal (showing the message sent)
	# RETURNS: nothing

	if ($script:post_w_creds_flg) {
		$uri = "http://$bi_webserver/admin?log=$msg&level=$level&user=$script:bi_username&pw=$script:bi_pw"
	} else {
		$uri = "http://$bi_webserver/admin?log=$msg&level=$level"
	}
	$result = Invoke-WebRequest -Uri $uri
	
	# @@@ DEV - RETAIN FOR TESTING/DEBUGGING
	#Write-Host "uri = $uri"
	#Write-Host "result = $result"	

	# HANDLE NOTIFICATIONS
	if ($silent_flg -ne $true) {
		msg.exe * /TIME:10 "Custom BI logfile message posted...`nlevel = $level`n$msg"
	}
	Write-Host "`nCustom BI logfile message posted: $msg"
}

function SendPushover($msg_str, $priority_level, $silent_flg) {
	# PURPOSE: Send custom notification via the Pushover API
	# ARGS:    $msg            = (string)
	#          $priority_level = (integer) -2|-1|0|1|2; -2=lowest, -1=low(silent), 0=normal, 1=high-priority, 2=emergency
	#          $silent_flg     = (boolean) $true prevents default output to terminal (showing the message sent)
	# RETURNS: nothing

	$uri = 'https://api.pushover.net/1/messages.json'
	$parameters = @{
		token    = $script:po_token
		user     = $script:po_user
		priority = $priority_level
		retry    = $script:po_retry
		expire   = $script:po_expire
		sound    = $script:po_sound
		device   = $script:po_device
		message  = $msg_str
	}
	#$parameters | Invoke-RestMethod -Uri $uri -Method Post
	$parameters | Invoke-RestMethod -Uri $uri -Method Post | Out-Null

	# PROVIDE FEEDBACK
	if ($silent_flg -ne $true) {
		msg.exe * /TIME:10 "Pushover notification sent...`npriority_level = $priority_level`n$msg_str"
	}
	Write-Host "`nPushover notification sent...`n$msg"
}

# SET DEFAULT USER SETTINGS
$variable_test = Get-Variable -Name reboot_flg -Scope Local -ErrorAction SilentlyContinue
if ($variable_test -eq $null) {
	$reboot_flg = $false
}
$variable_test = Get-Variable -Name post_to_BI_logfile_flg -Scope Local -ErrorAction SilentlyContinue
if ($variable_test -eq $null) {
	$post_to_BI_logfile_flg = $false
}
$variable_test = Get-Variable -Name send_po_notification_flg -Scope Local -ErrorAction SilentlyContinue
if ($variable_test -eq $null) {
	$send_po_notification_flg = $false
}

# INITIALIZE
$script_name = $MyInvocation.MyCommand.Name
$service_name = "BlueIris"

Write-Host "========================================================================="
Write-Host "Blue Iris Memory Leak Fix Utility"
Write-Host "-------------------------------------------------------------------------"
Write-Host "Note: use '$script_name demo' to prevent actions & notifications"
Write-Host "Note: use '$script_name restart|reboot' to force an action"
Write-Host "-------------------------------------------------------------------------"
Write-Host "source file = $path\$fname"

# HANDLE OPTIONAL COMMANDLINE ARGS
$demo_flg = $false
if ($p1) {
	if ($p1 -eq 'demo') {
		$demo_flg = $true
	}
	if ($p1 -eq 'restart') {
		Get-Service -Name $service_name
		Restart-Service -Name $service_name
		exit
	}
	if ($p1 -eq 'reboot') {
		Restart-Computer
		exit
	}
}

# ABORT IF LOG FILE NOT FOUND
if ( -not(Test-Path -LiteralPath "$path\$fname") ) {
	Write-Host "Sorry - source file not found!`n"
	Write-Host ""
	exit
}

# IMPORT PREVIOUS ACTION COUNTER FROM TXT FILE
$action_cntr_filespec = "$path\memleak_action_cntr.txt"
#Write-Host "action_cnts file = $action_cntr_filespec"
if (Test-Path "$action_cntr_filespec") {
	$action_cntr = Get-Content -Path "$action_cntr_filespec"
	$action_cntr = [int]$action_cntr
} else {
	$action_cntr = 0
}

# INITIALIZE VARS
$N = 4  # default number of data rows to extract
if ( ($ram_delta_threshold_MB -eq 0) -and ($ram_ratio_threshold -eq 0) ) {
	$ram_delta_threshold_MB = 50
}
$method = "deltas"
$method_threshold = $ram_delta_threshold_MB
$use_ram_ratio_flg = $false
$use_ram_step_flg = $false

# DETERMINE METHODS BASED ON USER SETTINGS
if ( ($ram_delta_threshold_MB -eq 0) -and ($ram_ratio_threshold > 1) ) {
	$use_ram_ratio_flg = $true
	$method = "ratios"
}
if ($ram_delta_threshold_MB -gt 0) {
	$use_ram_step_flg = $true
}

# GET THE LAST N ROWS FROM THE SOURCE FILE
# ... Note: LiteralPath allows use of square brackets (wildcard chars) in filespec; 
#     REF https://stackoverflow.com/a/31976040/13499984
$csvdata = Import-Csv -LiteralPath "$path\$fname"
$lastRows = $csvdata | Select-Object -Last $N
[array]::Reverse($lastRows)
if ($lastRows.Count -lt $N) {
	Write-Host "Sorry - source file must contain at least 4 data rows!`n"
	exit
}

Write-Host "action_cntr BEFORE check = $action_cntr"

$used_bytes = [System.Collections.ArrayList]@()  # initialize collector
foreach ($row in $lastRows) {
	#Write-Host "$row"
	$bytes = $($row.SoftwareUsedRamBytes)
	$used_bytes.Add([decimal]$bytes) | Out-Null
}
#Write-Host "use_ram_ratio_flg = $use_ram_ratio_flg"
#Write-Host "use_ram_step_flg = $use_ram_step_flg"

# CALCULATE RAM DELTAS &  RATIOS
$deltas = [System.Collections.ArrayList]@()  # initialize collector
$ratios = [System.Collections.ArrayList]@()  # initialize collector
$i = 0
foreach ($bytes_val in $used_bytes) {
	#Write-Host "bytes = $bytes_val"
	if ($i -gt 0) {
		if ($i -lt $used_bytes.Count) {
			$delta = $prev_val - $bytes_val
			$val = ([decimal]$delta/1MB).toString("0")
			$deltas.Add([int]$val) | Out-Null
			$ratio = $prev_val/$bytes_val
			$val = ([decimal]$ratio).toString("0.000000")
			$ratios.Add([decimal]$val) | Out-Null
		}
	}
	$prev_val = $bytes_val
	$i++
}

# TEST FOR MEMORY EVENT
# ... NOTE: A CONT-LEAK EVENT TAKES PRECEDENCE OVER A RAM-STEP EVENT
$event_found_flg = $true  # ASSUME A LEAK HAS OCCURRED
$event_type = ""
$msg  = ""
$msg1 = "Event type: CONT-LEAK (newest 3 " + $method + " exceed '" + $method_threshold + " MB')"
$msg2 = "Event type: RAM_STEP (ratio3 exceeds '" + $step_ratio_threshold + "' AND ratio1 & ratio2 exceed '" + $step_plateau_ratio_threshold + "')"

# 1) CONT-LEAK EVENT 
#    CRITERION: MOST RECENT 3 RATIOS MUST EQUAL OR EXCEED THRESHOLD VALUE
if ($use_ram_ratio_flg) {
	for ($i=0; $i -le 2; $i++) {
		$ratio = $ratios[$i]
		if ($ratio -lt $ram_ratio_threshold) {
			$event_found_flg = $false
			break
		}	
	}
} else {
	for ($i=0; $i -le 2; $i++) {
		$delta_MB = $deltas[$i]
		if ($delta_MB -lt $ram_delta_threshold_MB) {
			$event_found_flg = $false
			break
		}	
	}
}
#Write-Host "RAM event_found_flg = $event_found_flg"

# 2) RAM-STEP EVENT
#     CRITERIA: OLDEST used_bytes value must exceed $step_ram_min
#               RATIO #3 MUST EXCEED $step_ratio_threshold, AND
#               RATIOS #1,#2 MUST EXCEED $step_plateau_ratio_threshold
if ($event_found_flg -eq $false) {
	if ($use_ram_step_flg -eq $true) {
		if ($used_bytes[3] -ge $step_ram_min) {
			if ($ratios[2] -ge $step_ratio_threshold) {
				$test1 = ( ($ratios[0] -gt $step_plateau_ratio_threshold) )
				$test2 = ( ($ratios[1] -gt $step_plateau_ratio_threshold) )
				if ($test1 -and $test2) {
					$event_type = "step_type"
					$event_found_flg = $true
					$msg = $msg2
					if ($act_after_N_warnings > 0) {
						$act_after_N_warnings = 1
					}
				}
			}
		}
	}
} else {
	$event_type = "continuous_type"
	$msg = $msg1
}

# @@@ DEV - RETAIN FOR TESTING/DEBUGGING
#$event_found_flg = $true
#$msg = "RAM event simulated"

# UPDATE RESTART COUNTER
if ($event_found_flg) {
	$action_cntr++
} else {
	$action_cntr = 0  # Reset the counter	
}

# SHOW TEST RESULTS
Write-Host "latest $N log entries: (newest to oldest)"
Write-Host "bytes      = $used_bytes"
Write-Host "ratios     = $ratios"
Write-Host "deltas(MB) = $deltas"
if ($event_type -eq "step_type") {
	Write-Host("step ratio = " + $ratios[2] + "; plateau ratios: " + $ratios[0] + ", " + $ratios[1])
} else {
	if ($used_bytes[3] -lt $step_ram_min) {
		Write-Host "oldest bytes value < $step_ram_min; step-type event check skipped"
	}		
}
if ($event_found_flg) {	
	Write-Host "`n>> RAM EVENT DETECTED <<"
	if ($msg -ne "") {
		Write-Host "$msg"
	}
} else {
	Write-Host "`n>> NO EVENT DETECTED <<"	
}
Write-Host "`naction_cntr AFTER check = $action_cntr"

# HANDLE NOTIFICATIONS
if ($event_found_flg) {	
	if (!$demo_flg) {
		# POST CUSTOM MESSAGE TO BLUE IRIS LOGFILE (IF REQUESTED)
		if ($post_to_BI_logfile_flg) {
			$msg = "memory%20leak%20$event_type%20detected!%20($script_name)"
			$level = 0
			PostCustomBiLogEntry $msg $level $true
		}
		# PUSH MESSAGE VIA PUSHOVER (IF REQUESTED)
		# ... BUILD THE MESSAGE
		$hhmm = Get-Date -Format HH:mm
		$msg = "HEADS UP: memleak $event_type detected (" + $hhmm + ")"
		$msg1 = "Recent 4 vals: "
		$i = 0
		foreach ($val in $used_bytes) {
			$disp = [math]::Round($val/1GB,3)
			$disp = [string]$disp
			$disp = $disp.PadRight(5,'0')
			$z = [char](65 + $script:i)
			$msg1 += "$z=$disp" + "G, "
			$i++
		}
		$msg2 = "Ratios: "
		$i = 0
		foreach ($val in $ratios) {
			$disp = [math]::Round($val,4)
			$disp = [string]$disp
			$disp = $disp.PadRight(5,'0')
			$c = 65 + $script:i
			$z1 = [char](65 + $script:i)
			$z2 = [char](65 + $script:i + 1)
			$msg2 += "$z1/$z2=$disp" + ", "
			$i++
		}
		if ($event_type -ne "step_type") {
			$msg3 = "`nDeltas: "
			$i = 0
			foreach ($val in $deltas) {
				$disp = [string]$val
				$c = 65 + $script:i
				$z1 = [char](65 + $script:i)
				$z2 = [char](65 + $script:i + 1)
				$msg3 += "$z1-$z2=$disp" + "M, "
				$i++
			}
			$msg4 = ""
		} else {
			$msg3 = ""
			$msg4 = "`nStep-ratio: " + [math]::Round($ratios[2],4) + "; plateau_ratios: " + [math]::Round($ratios[0],4) + ", " + [math]::Round($ratios[1],4)
		}
		$msg1 = $msg1.TrimEnd(', ')
		$msg2 = $msg2.TrimEnd(', ')
		$msg3 = $msg3.TrimEnd(', ')
		$msg += "`n$script:msg1`n$script:msg2$script:msg3$script:msg4"

		if ($act_after_N_warnings -gt 0) {
			$x = $act_after_N_warnings-$action_cntr
			if ($reboot_flg) {
				$str1 = "(BI server is being rebooted now)"
				$str2 = "(BI server will be rebooted after $script:x more warnings)"
			} else {
				$str1 = "(BI service is being restarted now)"
				$str2 = "(BI service will be auto-restarted after $script:x more warnings)"
			}
			if ($action_cntr -ge $act_after_N_warnings) {
				$msg = "$script:msg`n$script:str1"
			} else {
				$msg = "$script:msg`n$script:str2"
			}
		} else {
			$msg = "`n$script:msg`n(consider restarting Blue Iris service)"
		}
			
		# ... SEND NOTIFICATION
		if ($send_po_notification_flg) {
			$hhmm = Get-Date -Format HH:mm
			SendPushover $msg $po_priority $true
		}
	}
}

# TAKE ACTION
if ($act_after_N_warnings -gt 0) {  # zero is an opt-out
	if (!$demo_flg) {
		if ($event_found_flg) {
			if ($action_cntr -ge $act_after_N_warnings) {
				if ($reboot_flg) {
					# REBOOT BLUE IRIS SERVER (IF REQUESTED)
					Restart-Computer
				} else { 
					# RESTART BLUE IRIS SERVICE (DEFAULT)
					Get-Service -Name $service_name
					Restart-Service -Name $service_name
				}
				$action_cntr = 0
			}
		}
		# UPDATE ACTION COUNTER TO TXT FILE
		$action_cntr | Set-Content "$action_cntr_filespec"
	}
}

# DEMO MODE HEADS UP
if ($action_cntr -gt 0) {
	if ($demo_flg) {
		Write-Host "`nDemo mode:`n-No actions taken"
		Write-Host "-action_cntr reset to 0"
		$action_cntr = 0
	}
}

Write-Host ""
